// -*- C++ -*- std header                                                      |
//-----------------------------------------------------------------------------+
//                                                                             |
// yanc.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2002                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//                        YET ANOTHER N-BODY CODE                              |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class yanc                         see yanc.cc                              |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef included_yanc_h
#define included_yanc_h

#ifndef included_iostream
#  include <iostream>
#  define included_iostream
#endif
using std::ostream;

#ifndef included_nbio_h
#  include <nbdy/nbio.h>
#endif

#ifndef included_drct_h
#  include <nbdy/drct.h>
#endif
////////////////////////////////////////////////////////////////////////////////
namespace nbdy {                               // yanc is in namespace nbdy     
  class ndata;                                 // forward declare ndata         
  class bodies;                                // forward declare bodies        
  class basic_nbody;                           // forward declare nbody         
  class potential;                             // forward declare potential     
  //----------------------------------------------------------------------------
  class pot_provider {
    public: virtual potential* get_pot(const char*) = 0;
  };
  //----------------------------------------------------------------------------
  class yanc {
  private:
    yanc();                                    // not implemented               
    yanc(const yanc&);                         // not implemented               
    yanc& operator= (const yanc&);             // not implemented               
  protected:
    ndata       *MYNDATA;                      // our N-body data               
    basic_nbody *MYNBODY;                      // our N-body code               
    const bodies*mybodies() const;             // our N bodies                  
    const char*  pot_file() const;             // our file for external pot     
    const double r_max   () const;             // max r for bodies in tree      
  public:
    //--------------------------------------------------------------------------
    yanc(                                      // constructor without parameters
	 const char*,                          // I: yanc-formatted input file  
	 pot_provider* =0);                    //[I: provider for external pot] 
#ifdef ALLOW_NEMO
    //--------------------------------------------------------------------------
    yanc(                                      // constructor taking parameters 
	 const char*,                          // I: input file (yanc/nemo)     
	 const bool,                           // I: do nemo I/O?               
	 const double,                         // I: theta                      
	 const int,                            // I: Nreuse                     
	 const int,                            // I: Ncrit                      
	 const double,                         // I: eps                        
	 const int,                            // I: kern                       
	 const int,                            // I: hmin                       
	 const int,                            // I: Nsteps                     
	 const double,                         // I: f_a                        
	 const double     = 0.,                //[I: f_p]                       
	 const double     = 0.,                //[I: f_c]                       
	 const int        = 16u,               //[I: Nref]                      
	 const bool       = false,             //[I: resume old (if nemo I/O)   
	 const potential* = 0,                 //[I: external potential]        
	 const double     = 0.0,               //[I: r_max]                     
         const bool       = false,             //[I: read aux data?]            
	 const int        = 2,                 //[I: # nemo output streams]     
	 const unsigned[4]= nbdy::DIRECT       //[I: direct sum control]        
#ifdef __GRAPE
	 ,const bool      = false              //[I: use GRAPE?]                
#endif
	 );
    //--------------------------------------------------------------------------
    void  describe_nemo   (                    // describe simulation           
			   ostream&,           // I: output stream              
			   const char*);       // I: command line               
    //--------------------------------------------------------------------------
    void  open_nemo       (                    // open NEMO output stream       
			   const int   =0,     //[I: index of nemo stream]      
			   const char* =0,     //[I: file name ]                
			   const bool  =false);//[I" resume old sim?]           
    //--------------------------------------------------------------------------
    void  close_nemo      (                    // close NEMO output stream      
			   const int =0);      //[I: index of nemo stream]      
    //--------------------------------------------------------------------------
    void  write_nemo      (                    // NEMO-format output            
			   const io  =io::mxv, //[I: what to write out]         
			   const int =0,       //[I: index of nemo stream]      
			   const bool=true);   //[I: output diagnostics?]       
    //--------------------------------------------------------------------------
    int   nemo_devices    () const;            // R: # nemo output streams      
    //--------------------------------------------------------------------------
    bool  nemo_is_open    (                    // ready for NEMO output ?       
			   const int =0) const;//[I: index of nemo stream]      
    //--------------------------------------------------------------------------
#endif // ALLOW_NEMO
    //--------------------------------------------------------------------------
    ~yanc();                                   // destructor                    
    //--------------------------------------------------------------------------
    void  full_step       ();                  // perform one full time step    
    //--------------------------------------------------------------------------
    void  describe        (                    // describe simulation           
			   ostream&,           // I: output stream              
			   const char* =0,     //[I: name of main]              
			   const char* =0);    //[I: base name: output files]   
    //--------------------------------------------------------------------------
    static
    void  describe_yanc_format(                // describe yanc I/O format      
			       ostream&,       // I/O: stream to write to       
			       const bool=false);
    //--------------------------------------------------------------------------
    bool  write_yanc_ascii(                    // YANC-format output to file    
			   const char*,        // I: file name                  
			   const char* =0);    //[I: name of main ]             
    //--------------------------------------------------------------------------
    bool  write_yanc_binary(                   // write yanc form, bodies binary
			    const char*,       // I:   name of output file      
			    const char* =0)const;// I: name of calling program  
    //--------------------------------------------------------------------------
    void  long_stats      (ostream&) const;    // write long stats              
    void  short_stats     (ostream&) const;    // write short stats             
    void  short_head      (ostream&) const;    // write header for short stats  
    bool  okay            ()         const;    // is all fine?                  
    int   Nsteps          ()         const;    // # time steps levels           
    int   Nbodies         ()         const;    // # bodies                      
    const char* data_file ()         const;    // input data file               
    float eps             ()         const;    // softening length              
    float tau_min         ()         const;    // minimum time step             
    float tau_max         ()         const;    // maximum time step             
    int   softening       ()         const;    // type of softening             
    int   kernel          ()         const;    // type of kernel                
    float time            ()         const;    // current simulation time       
    float initial_time    ()         const;    // time at input                 
    io    writing         ()         const;    // what to write out by default  
    float theta           ()         const;    // opening parameter             
    int   Ncrit           ()         const;    // # max bodies / un-split cell  
    int   Nreuse          ()         const;    // # max subsequent tree re-uses 
    int   Nref            ()         const;    // # bodies in cell for rho estim
    float facc            ()         const;    // time-stepping parameter       
    float fpot            ()         const;    // time-stepping parameter       
    float fcom            ()         const;    // time-stepping parameter       
    float cpu_total       ()         const;    // accumulated cpu time [sec]    
    //--------------------------------------------------------------------------
  };
}
//------------------------------------------------------------------------------
#endif // included_yanc_h
